/* lqphrase.c -- Copyright 1989, 1990 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 *
 * $Id: lqphrase.c,v 1.6 93/10/24 14:30:42 lee Exp $
 */

/* lqphrase, part of Liam Quin's text retrieval package...
 *
 * lqphrase is intended to be an example of one way to use the programming
 * interface to lq-text.
 *
 * The idea is quite simple:
 * Simply take a phrase p, held in a string (char *p), and call
 *	t_Phrase *Phrase = String2Phrase(p);
 * The result, if not null, contains only one interesting thing at this
 * point:
 *	Phrase->ModifiedString
 * is the canonical version of p -- with common and short words removed.
 * for example,
 *	p = "The boy sat down in His Boat and playd with his toes.";
 * might result in Phrase->ModifiedString containing
 *	"[*the] boy sat down [in] [*his] boat [*and] [?playd] with [*his] toe"
 * Common words are marked with a *, and unknown words with ?.
 * [NOTE: modifiedPhrase is no longer be supported]
 *
 * An attempt may have been made to reduce plurals.
 *
 * Since this phrase contains a word not in the database (playd), it will
 * never match anything.  As a result, it is a good idea to print this string
 * (possibly massaging it first) so users can see what is going on.  If you
 * have it, the curses-based "lqtext" does this.
 *
 * If we change "playd" to "played", the above string is equivalent to
 *	"[*the] boy sat down [xx] [*the] boat [*the] played with [*the] toe"
 * In other words, all common words are equivalent.  The package remembers
 * that one or more common words were skipped, and also that one or more
 * lumps of letters too small to make up a word were skipped.
 * Compare also h/wordrules.h for more info, and liblqtext/Phrase.c for
 * the code that actually does the matching.
 *
 * Now, having sorted that out, we have our canonical string (and lots of
 * other things) in Phrase, so we can now call
 *	MakeMatches(Phrase);
 * This will return the number of matches (*NOT* the number of files) for
 * the given ModifiedPhrase in the database.
 * This can take several seconds, so again, it can be worth printing out
 * the modified string as soon as it is available, so the user is looking at
 * that whilst MakeMatches is working!  I have experimented with faster
 * versions of MakeMatches involving binary search, but the extra complexity
 * slowed things down on smaller databases.  I don't have enough disk space
 * here to make a large enough database to do real timings, sorry.
 *
 * Now we have done MakeMatches, we can march along the linked list of
 * pointers to linked lists of arrays of matches.  Clear?  No?  Well,
 * that's why there's an example.  See Match() below.
 *
 * Now, each match currently gives us
 * t_FID FID; Files are numbered from 1 in the database
 * unsigned long BlockInFile; -- the block in the file
 * unsigned char WordInBlock; -- the word in the block
 * unsigned char StuffBefore; -- the amount of leading garbage
 * unsigned char Flags, including (see wordrules.h):
 *
 * WPF_WASPLURAL		The word...  ended in s
 * WPF_UPPERCASE		...Started with a capital letter
 * WPF_POSSESSIVE		...ended in 's
 * WPF_ENDEDINING		...ended in ing
 * WPF_LASTWASCOMMON	the previous word was common
 * WPF_LASTHADLETTERS	we skipped some letters to get here
 * WPF_LASTINBLOCK	I'm the last word in this block
 *
 */

#include "error.h"
#include "globals.h" /* defines and declarations for database filenames */

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <fcntl.h>
#include <sys/types.h>
#include <malloc.h>
#include "emalloc.h"
#include "fileinfo.h" /* for wordinfo.h */
#include "wordinfo.h"
#include "pblock.h"
#include "phrase.h"
#include "readfile.h"

#ifndef STREQ
# define STREQ(boy,girl) ((*(boy) == *(girl)) && (!strcmp((boy),(girl))))
#endif

extern int AsciiTrace;
extern t_PhraseCaseMatch PhraseMatchLevel;

/** System calls and functions... **/
/** Unix system calls used in this file: **/
extern void exit();

/** Unix Library Functions used: **/
/** lqtext library functions: **/
extern void SetDefaults();
extern void DefaultUsage();

/** functions used before they're defined within this file: **/
void Match();
/** **/

static char *Revision = "@(#) $Id: lqphrase.c,v 1.6 93/10/24 14:30:42 lee Exp $";

char *progname = "lqphrase";

int SilentMode = 0; /* don't print matches if set to one */
int PrintNumberOfWordsInMatchedPhrases = 1;

int
main(argc, argv)
    int argc;
    char *argv[];
{
    extern int optind, getopt();
    extern char *optarg;
    int ch;
    int ErrorFlag = 0;
    char *InputFile = 0;

    progname = argv[0];

    SetDefaults(argc, argv);

    while ((ch = getopt(argc, argv, "Zz:af:hNpr:slxVv")) != EOF) {
	switch (ch) {
	case 'z':
	case 'Z':
	    break; /* done by SetDefaults(); */
	case 'V':
	    fprintf(stderr, "%s version %s\n", progname, Revision);
	    break;
	case 'v': /* same as -t 1 */
	    AsciiTrace++;
	    break;
	case 'f':
	    InputFile = optarg;
	    break;
	case 'l':
	    break; /* list mode is the default */
	case 'N':
	    PrintNumberOfWordsInMatchedPhrases = 0;
	    break;
	case 'r':
	    /* ignored for compat. with lqrank*/
	    break;
	case 's':
	    SilentMode = 1;
	    break;
	case 'x':
	    ErrorFlag = (-1);
	    break;
	case '?':
	    ErrorFlag = 1;
	}
    }

    if (ErrorFlag) {
	fprintf(stderr, "Usage: %s [options] \"phrase\" [...]\n", progname);
	fprintf(stderr, "%s: options are:\n", progname);
	fputs("\
	-l	-- list mode, suitable for lqshow (the default)\n\
	-s	-- silent mode; exit status indicates success of matching\n\
\n", stderr);

	DefaultUsage();
	exit( ErrorFlag > 0 ? 1 : 0); /* 0 means -x was used */
    }
    
    if (AsciiTrace > 1) {
	switch (PhraseMatchLevel) {
	case PCM_HalfCase:
	    fprintf(stderr, "%s: Matching phrases heuristically.\n", progname);
	    break;
	case PCM_SameCase:
	    fprintf(stderr, "%s: Matching phrases precisely.\n", progname);
	    break;
	case PCM_AnyCase:
	    fprintf(stderr, "%s: Matching phrases approximately.\n", progname);
	    break;
	default:
	    fprintf(stderr, "%s: internall error, case matching is %d\n",
						progname, PhraseMatchLevel);
	    exit(2);
	}
    }

    if (InputFile) {
	FILE *f;
	char NeedClose = 1;
	char *theLine;

	if (STREQ(InputFile, "-")) {
	    f = stdin;
	    NeedClose = 0;
	} else {
	    extern FILE *fEopen();

	    f = fEopen(E_FATAL|E_SYS, InputFile, "List of phrases", "r");
	}

	while (fReadLine(f, &theLine, UF_NORMAL) >= 0) {
	    if (theLine && *theLine) {
		Match(theLine);
	    }
	    if (theLine) {
		(void) efree(theLine);
	    }
	}

	if (NeedClose) {
	    (void) close(f);
	}
    }

    while (optind < argc) {
	Match(argv[optind++]);
    }

    if (SilentMode) {
	/* if we got to here we didn't find anything */
	exit(1);
    }
    return 0;
}

static int NumberOfWords = 1;

static int
PrintAndAcceptOneMatch(Phrase, Match)
    t_Phrase *Phrase;
    t_Match *Match;
{
    extern t_FileInfo *GetFileInfo();
    static t_FID LastFID = (t_FID) 0;
    static t_FileInfo *FileInfo = 0;

    if (Match != (t_Match *) 0) {
	if (Match->Where->FID != LastFID) {
	    t_FID FID = Match->Where->FID;
	    /*TODO: use DestroyFileInfo instead of efree:... */

	    if (FileInfo) {
		efree((char *) FileInfo);
	    }

	    if ((FileInfo = GetFileInfo(FID)) == (t_FileInfo *) 0) {
		return 0; /* don't want this match! */
	    }
	    LastFID = FID;
	}

	/* Now that we know that we have something to print... */
	if (SilentMode) {
	    exit(0); /* OK, found something */
	}
	if (AsciiTrace) {
	    printf(
		"NW %-3d BIF %-7lu WIB %-7u Sep %-3d F %-04o FID %-4d %s\n",
		NumberOfWords,
		Match->Where->BlockInFile,
		(unsigned) Match->Where->WordInBlock,
		(unsigned) Match->Where->StuffBefore,
		(unsigned) Match->Where->Flags,
		FileInfo->FID,
		FileInfo->Name
	    );
	} else {
	    if (PrintNumberOfWordsInMatchedPhrases) {
		printf("%d %lu %u %lu %s\n",
		    NumberOfWords,
		    Match->Where->BlockInFile,
		    Match->Where->WordInBlock,
		    FileInfo->FID,
		    FileInfo->Name
		);
	    } else {
		printf("%lu %u %lu %s\n",
		    Match->Where->BlockInFile,
		    Match->Where->WordInBlock,
		    FileInfo->FID,
		    FileInfo->Name
		);
	    }
	}
    }
    return 0;
}

void
Match(Phrase)
    char *Phrase;
{
    extern t_Phrase *String2Phrase();
    extern long MakeMatchesWhere();

    t_Phrase *P;
    t_MatchList *Matches;

    if (!Phrase || !*Phrase) return;
    if ((P = String2Phrase(Phrase)) == (t_Phrase *) 0) return;

    /* Count words in the phrase: */
    if (PrintNumberOfWordsInMatchedPhrases || AsciiTrace) {
	/* NumberOfWords is global to this file, so that the
	 * callback PrintAndAcceptOneMatch() can use it:
	 */
	NumberOfWords = NumberOfWordsInPhrase(P);
    }

    if (MakeMatchesWhere(P, PrintAndAcceptOneMatch) <= 0L) return;

#if 0
    if (P) {
	for (Matches = P->Matches; Matches != (t_MatchList *) 0;
						Matches = Matches->Next) {
	    (void) PrintAndAcceptOneMatch(P, Matches->Match);
	}
    }
#endif
}
