/* lqrank.c -- Copyright 1983, Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 *
 * $Id: lqrank.c,v 1.8 93/10/24 14:30:29 lee Exp $
 */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <malloc.h>
#include "emalloc.h"

#include <sys/types.h>
#include "readfile.h"
#include "fileinfo.h"
#include "wordinfo.h"
#include "phrase.h"
#include "sys/stat.h"

/** Macros and variable definitions **/

extern int AsciiTrace;

/** System calls and functions... **/
/** Unix system calls used in this file: **/
extern void exit();

/** Unix Library Functions used: **/
/** lqtext library functions: **/
extern void SetDefaults();
extern void DefaultUsage();

/** functions used before they're defined within this file: **/
void Match();
static void AddMatch(/*Fp, FileName, FID, mp */);
/** **/

static char *Revision = "@(#) $Id: lqrank.c,v 1.8 93/10/24 14:30:29 lee Exp $";

char *progname = "lqrank";

typedef struct s_RMatches {
    long BlockInFile;
    int WordInBlock;
    int WordsInPhrase;
    double MatchRank;
    int WhichPhrase;
    struct s_RMatches *Next;
} t_Matches;

typedef struct s_RMatchList {
    char *Name;
    t_FID FID;
    t_Matches *Matches;
    struct s_RMatchList *Next;
} t_RMatchList;

typedef struct s_File {
    char *Name;
    t_FID FID;
    t_RMatchList *MatchList;
    double FileRank;
    struct s_File *Next;
} t_File;

t_File *Files = 0;

typedef enum {
    e_Any, e_Most, e_All
} t_RankType;

static t_RankType RankType = e_All;

static void SetRankingMode();
static void PrintResults();
static int AddFileName();
static int AddFile();
static void AddPhrase();
static int SortFlag = 1; /* default is to sort by filename within a rank */
static int OneOnlyFlag = 0; /* print only one match per document */
static void ReadMatchesFromFile();
static int FileCount = 0;
static int FailThreshold = 0;
static int OnlyOnePhraseAnywaySoWhyBother = 0;

int
main(argc, argv)
    int argc;
    char *argv[];
{
    extern int optind, getopt();
    extern char *optarg;
    int ch;
    int ErrorFlag = 0;

    progname = argv[0];

    SetDefaults(argc, argv);

    while ((ch = getopt(argc, argv, "Zz:f:op:r:uxVv")) != EOF) {
	switch (ch) {
	case 'z':
	case 'Z':
	    break; /* done by SetDefaults(); */
	case 'V':
	    fprintf(stderr, "%s version %s\n", progname, Revision);
	    break;
	case 'v': /* same as -t 1 */
	    AsciiTrace++;
	    break;

	case 'f':
	    if (AddFileName(optarg) < 0) {
		exit(1);
	    }
	    break;
	case 'o':
	    OneOnlyFlag = 1;
	    break;
	case 'p':
	    AddPhrase(optarg);
	    break;
	case 'r':
	    SetRankingMode(optarg);
	    break;
	case 'u':
	    SortFlag = 0;
	    break;
	case 'x':
	    ErrorFlag = (-1);
	    break;
	case '?':
	    ErrorFlag = 1;
	}
    }

    /* Normally put call to lrqError here to give a helpful message,
     * but not yet ready to ship the error handling package, sorry
     */
    if (ErrorFlag) {
	fprintf(stderr, "Usage: %s [options]\n", progname);
	fprintf(stderr, "%s: options are:\n", progname);
	fputs(" -o     	-- print only one match from each document\n", stderr);
	fputs(" -p p	-- look up p as a phrase with lqphrase(1)\n", stderr);
	fputs(" -r rank	-- set ranking level for subsequent files\n", stderr);
	if (AsciiTrace) {
	    fprintf(stderr, "\trank can be none, most or all.\n");
	}
	fputs(" -u 	-- produce unsorted output\n", stderr);
	fputs(" -f file	-- file contains list of matches\n", stderr);
	DefaultUsage();
	fputs("Warning: all -f options should normally come last!\n", stderr);
	exit( ErrorFlag > 0 ? 1 : 0); /* 0 means -x was used */
    }
    
    if (argc - optind  == 1 && !STREQ(argv[optind], "-")) {
	OnlyOnePhraseAnywaySoWhyBother = 1;
    }

    if (optind < argc) {
	for (; optind < argc; ++optind) {
	    if (STREQ(argv[optind], "-")) {
		char *theLine = 0;
		while (fReadLine(
				stdin,
				&theLine,
				UF_IGNBLANKS|UF_IGNSPACES|UF_IGNHASH) >= 0) {
		    AddPhrase(theLine);
		    (void) efree(theLine);
		    theLine = 0;
		}
		if (theLine) {
		    (void) efree(theLine);
		}
	    }
	    (void) AddPhrase(argv[optind]);
	}
    }

    if (FileCount) {
	PrintResults();
	exit(0);
    } else {
	exit(1);
    }
}

static char *
PhraseMatchString()
{
    extern t_PhraseCaseMatch PhraseMatchLevel;

    switch (PhraseMatchLevel) {
    case PCM_SameCase:
	return "p";
    case PCM_HalfCase:
    default:
	return "h";
    case PCM_AnyCase:
	return "a";
    }
}

static char *
AppendQuotedString(Dest, in)
    char *Dest;
    char *in;
{
    register char *q = Dest;
    register char *p = in;

    *q++ = '"';

    for (; *p; p++) {
	switch (*p) {
	case '\n': case '\f': case '\t': case ' ':
	    *q++ = ' ';
	    break;
	case '"': case '\\': case '$': case '`':
	    *q++ = '\\';
	    *q++ = *p;
	    break;
	default:
	    *q++ = *p;
	}
    }

    *q++ = '"';
    *q = '\0';
    return q;
}

static int NumberOfWords = 1;
static int MatchesSeenInThisFIle = 0;

static int
PrintAndAcceptOneMatch(Phrase, Match)
    t_Phrase *Phrase;
    t_Match *Match;
{
    extern t_FileInfo *GetFileInfo();
    static t_FID LastFID = (t_FID) 0;
    static t_FileInfo *FileInfo = 0;

    if (Match != (t_Match *) 0) {
	if (Match->Where->FID != LastFID) {
	    t_FID FID = Match->Where->FID;
	    /*TODO: use DestroyFileInfo instead of efree:... */

	    if (FileInfo) {
		efree((char *) FileInfo);
	    }

	    if ((FileInfo = GetFileInfo(FID)) == (t_FileInfo *) 0) {
		return 0; /* don't want this match! */
	    }
	    LastFID = FID;
	    if (OneOnlyFlag) {
		goto PrintIt;
	    }
	}

	if (!OneOnlyFlag) {
PrintIt:
	    printf("%d %lu %u %lu %s\n",
		NumberOfWords,
		Match->Where->BlockInFile,
		Match->Where->WordInBlock,
		FileInfo->FID,
		FileInfo->Name
	    );
	}
    }
    return 0;
}

static void
AddPhrase(Phrase)
    char *Phrase;
{
    extern t_Phrase *String2Phrase();
    extern t_FileInfo *GetFileInfo();
    extern long MakeMatches();

    t_Phrase *P;
    t_MatchList *Matches;
    t_FID LastFID = ~(t_FID) 0;
    t_FileInfo *FileInfo = 0;
    t_File **Fpp;
 
    if (!Phrase || !*Phrase) return;
 
    /* should we allow this, for "Most" (or statistical) ranking? */
    for (Fpp = &Files; *Fpp; Fpp = &(*Fpp)->Next) {
	if (STREQ((*Fpp)->Name, Phrase)) {
	    Error(E_WARN,
		"Phrase \"%s\" included more than once; duplicate ignored.",
		Phrase
	    );
	    return; /* not indexed */
	}
    }

    if ((P = String2Phrase(Phrase)) == (t_Phrase *) 0) return;
    NumberOfWords = NumberOfWordsInPhrase(P);
    if (OnlyOnePhraseAnywaySoWhyBother) {
	extern long *MakeMatchesWhere();

	(void) MakeMatchesWhere(P, PrintAndAcceptOneMatch);
    } else {
	if (MakeMatches(P) <= 0L) return;
    }

    /* Now Fpp points to the end of the list */

    *Fpp = (t_File *) emalloc("lqrank:AddPhrase.t_File", sizeof(t_File));
    (*Fpp)->Next = 0;
    (*Fpp)->Name = emalloc("lqrank:AddPhrase.File.Name", strlen(Phrase) + 1);
    (void) strcpy((*Fpp)->Name, Phrase);
    (*Fpp)->MatchList = (t_RMatchList *) 0;
    (*Fpp)->FileRank = 0;
    ++FileCount;

    for (Matches = P->Matches; Matches != (t_MatchList *) 0;
					    Matches = Matches->Next) {

	t_Matches *Mp;

	if (Matches->Match != (t_Match *) 0) {
	    if (Matches->Match->Where->FID != LastFID) {
		LastFID = Matches->Match->Where->FID;
		/*TODO: use DestroyFileInfo instead of efree:... */
		if (FileInfo) efree((char *) FileInfo);
		if ((FileInfo = GetFileInfo(LastFID)) == (t_FileInfo *) 0) {
		    continue;
		}
	    }
	    Mp = (t_Matches *) emalloc(
		"lqrank.AddFile.Matches",
		sizeof(t_Matches)
	    );

	    Mp->BlockInFile = Matches->Match->Where->BlockInFile;
	    Mp->WordInBlock = Matches->Match->Where->WordInBlock;
	    Mp->WordsInPhrase = NumberOfWords;
	    Mp->MatchRank = 0;
	    Mp->WhichPhrase = FileCount;;
	    AddMatch(*Fpp, FileInfo->Name, LastFID, Mp);
	}
    }
}

static int
AddFileName(Name)
    char *Name;
{
    int Result;
    FILE *f;
    extern int IsFile(), IsDir(), IsNZFile();
    extern FILE *fEopen();
    
    if (STREQ(Name, "-")) {
	return AddFile(stdin, "(standard input)");
    }
    /* The return value is the number of files indexed. */
    if (!IsFile(Name)) {
	if (IsDir(Name)) {
	    Error(E_WARN, "%s is a directory, not a file", Name);
	    return 0;
	}
	if (AsciiTrace) {
	    Error(E_WARN, "%s is not a regular file...", Name);
	}
    }

    f = fEopen(E_FATAL, Name, "match-list file", "r");
    if (!f) {
	/*NOTREACHED*/
	return -1;
    }
    Result = AddFile(f, Name);
    fclose(f);
    return Result;
}

static void
SetRankingMode(What)
    char *What;
{
    if (STREQ(What, "none") || STREQ(What, "or")) {
	RankType = e_Any;
    } else if (STREQ(What, "most") || STREQ(What, "quorum")) {
	RankType = e_Most;
    } else if (STREQ(What, "all") || STREQ(What, "and")) {
	RankType = e_All;
    } else {
	Error(E_FATAL|E_XHINT,
	    "-r: option must be \"none\", \"most\" or \"all\", not \"%s\"\n",
	    What ? What : ""
	);
    }
}

static int
AddFile(f, Name)
    FILE *f;
    char *Name;
{
    t_File **Fpp;

    for (Fpp = &Files; *Fpp; Fpp = &(*Fpp)->Next) {
	if (STREQ((*Fpp)->Name, Name)) {
	    Error(E_WARN,
		"File %s included more than once; duplicate ignored.",
		Name
	    );
	    return 0; /* not indexed */
	}
    }

    /* Now Fpp points to the end of the list */

    *Fpp = (t_File *) emalloc("lqrank:AddFile.File", sizeof(t_File));
    (*Fpp)->Next = 0;
    (*Fpp)->Name = emalloc("lqrank:AddFile.Name", strlen(Name) + 1);
    (void) strcpy((*Fpp)->Name, Name);
    (*Fpp)->MatchList = (t_RMatchList *) 0;
    ReadMatchesFromFile(*Fpp, f, Name);
    ++FileCount;
    return 1;
}

static void
AddMatch(Fp, FileName, FID, mp)
    t_File *Fp;
    char *FileName;
    t_FID FID;
    t_Matches *mp;
{
    static t_RMatchList **Mpp = 0;
    static t_Matches **mpp = 0;
    static t_File *LastFile = 0;
    register int i = 1;

    if (Fp != LastFile) {
	LastFile = Fp;
	Mpp = &(Fp->MatchList);
    } else {
	if (Mpp && *Mpp && (
	    ((*Mpp)->FID && (*Mpp)->FID != FID) ||
	    !STREQ((*Mpp)->Name, FileName)
	)) {
	    Mpp = &(Fp->MatchList);
	}
    }

    for (; *Mpp; Mpp = &(*Mpp)->Next) {
	if ((i = STRCMP((*Mpp)->Name, FileName)) >= 0) {
	    break;
	}
    }

    if (i != 0 || !*Mpp) {
	t_RMatchList *M = (*Mpp);
	*Mpp = (t_RMatchList *) emalloc(
	    "lqrank:AddMatch.Match",
	    sizeof(t_RMatchList)
	);
	(*Mpp)->Next = M;
	(*Mpp)->Name = emalloc("lqrank:AddMatch.FileName",strlen(FileName) + 1);
	(*Mpp)->FID = FID;
	(void) strcpy((*Mpp)->Name, FileName);
	(*Mpp)->Matches = 0;
    }

    /* Insert the match in sorted order */
    for (mpp = &(*Mpp)->Matches; *mpp; mpp = &(*mpp)->Next) {
	if ((*mpp)->BlockInFile > mp->BlockInFile) break;
	if ((*mpp)->BlockInFile == mp->BlockInFile) {
	    if ((*mpp)->WordInBlock > mp->WordInBlock) break;
	    if ((*mpp)->WordInBlock == mp->WordInBlock) {
		free(mp); /* already there! */
		return;
	    }
	}
    }

    /* insert */
    mp->Next = (*mpp);
    *mpp = mp;
}

static void
ReadMatchesFromFile(Fp, f, Name)
    t_File *Fp;
    FILE *f;
    char *Name;
{
    t_Matches *Mp;
    char *theLine;

    for (;;) {
	register char *p;
	t_FID FID;
	char *FileName;

	if (fReadLine(f, &theLine, UF_IGNBLANKS|UF_IGNSPACES|UF_IGNHASH) < 0) {
	    if (theLine) efree(theLine);
	    return;
	}

	if (RankType == e_Any && !SortFlag) {
	    puts(theLine);
	    efree(theLine);
	    continue;
	}

	/* fReadLine has swallowed any blank or commented lines,
	 * and has deleted any leading spaces.
	 */
	
	Mp = (t_Matches *) emalloc(
	    "lqrank:ReadMatchesFromFile.t_Matches",
	    sizeof(t_Matches)
	);
	Mp->Next = (t_Matches *) 0;
	Mp->BlockInFile = 0;
	Mp->WordInBlock = 0;
	Mp->WordsInPhrase = 0;
	Mp->MatchRank = 0;
	Mp->WhichPhrase = FileCount;

	/* don't use sscanf, for speed and for saving memory by keeping the
	 * filename in-place
	 */

	p = theLine;
	while (isspace(*p)) {
	    p++;
	}

	for (; isdigit(*p); p++) {
	    Mp->WordsInPhrase *= 10;
	    Mp->WordsInPhrase += *p - '0';
	}

	while (isspace(*p)) {
	    p++;
	}

	if (!isdigit(*p)) {
	    Error(E_WARN,
		"File %s contains bad match, expected digit in %s\n",
		theLine
	    );
	    efree(theLine);
	    return;
	}

	for (; isdigit(*p); p++) {
	    Mp->BlockInFile *= 10;
	    Mp->BlockInFile += *p - '0';
	}

	while (isspace(*p)) {
	    p++;
	}

	if (!isdigit(*p)) {
	    Error(E_WARN,
		"File %s conatins strange match, expected digit in %s\n",
		theLine
	    );
	    efree(theLine);
	    return;
	}

	for (; isdigit(*p); p++) {
	    Mp->WordInBlock *= 10;
	    Mp->WordInBlock += *p - '0';
	}

	while (isspace(*p)) {
	    p++;
	}

	/* FID */
	FID = 0L;

	while (isdigit(*p)) {
	    FID *= 10;
	    FID += *p - '0';
		p++;
	}

	while (isspace(*p)) {
	    p++;
	}

	if (!*p) {
	    static char *theName = 0;
	    static t_FID theFID = 0L;

	    t_FileInfo *FileInfo;
	    
	    if (FID == theFID) {
		FileName = theName;
	    } else {
		extern t_FileInfo *GetFileInfo();

		if (theName) {
		    efree(theName);
		}
		FileInfo = GetFileInfo(FID);
		theFID = FID;
		theName = FileName = FileInfo->Name;
		(void) efree(FileInfo);
	    }
	} else {
	    /* file name, already null-terminated */
	    FileName = p;
	}

	AddMatch(
	    Fp,
	    FileName,
	    FID,
	    Mp
	);
	efree(theLine);
    }
}

/* Keep a list of results for a given filename, so we can print them all
 * at once
 */
static t_Matches *FinalMatches = 0;

static void
PrintFinalResultsForOneFile(FileName, FID)
    char *FileName;
    t_FID FID;
{
    register t_Matches *Mp;

    for (Mp = FinalMatches; Mp; Mp = Mp->Next) {
	printf("%d %ld %ld %d %s\n",
	    Mp->WordsInPhrase,
	    Mp->BlockInFile,
	    Mp->WordInBlock,
	    FID,
	    FileName
	);
    }

    for (Mp = FinalMatches; Mp; ) {
	t_Matches *N = Mp->Next;
	free(Mp);
	Mp = N;
    }

    FinalMatches = 0;
}

static void
AddResult(WordsInPhrase, BlockInFile, WordInBlock, FID, Name)
    int WordsInPhrase;
    long BlockInFile;
    int WordInBlock;
    t_FID FID;
    char *Name;
{
    static char *FileName = 0;
    static t_FID LastFID = 0;
    register t_Matches **Mpp;
    register t_Matches *Mp;

    if (AsciiTrace) {
	fprintf(stderr, "+R %d %ld %d %s -> %s\n", WordsInPhrase, BlockInFile, WordInBlock, FileName ? FileName : "[NULL]", Name ? Name : "(FLUSH)");
    }

    if (!Name || !FileName || LastFID != FID || !STREQ(Name, FileName)) {

	/* deal with the cached results first: */
	if (FileName) {
	    PrintFinalResultsForOneFile(FileName, LastFID);
	    efree(FileName);
	    FileName = 0;
	}

	if (!Name) {
	    /* nothing more to do */
	    return;
	}

	/* now set up new cache: */
	FileName = emalloc("lqrank.ResultCache.Name", strlen(Name) + 1);
	(void) strcpy(FileName, Name);
	LastFID = FID;
    }

    for (Mpp = &(FinalMatches); *Mpp; Mpp = &(*Mpp)->Next) {
	if ((*Mpp)->BlockInFile > BlockInFile) break;
	if ((*Mpp)->BlockInFile == BlockInFile) {
	    if ((*Mpp)->WordInBlock == WordInBlock) {
		return; /* already there */
	    } else if ((*Mpp)->WordInBlock > WordInBlock) {
		break;
	    }
	}
    }

    Mp = *Mpp;
    *Mpp = (t_Matches *) emalloc("lqrank:ResultCache.Match",sizeof(t_Matches));
    (*Mpp)->Next = Mp;
    (*Mpp)->WordsInPhrase = WordsInPhrase;
    (*Mpp)->BlockInFile = BlockInFile;
    (*Mpp)->WordInBlock = WordInBlock;
}

static void PrintMatchesInAnyFiles();
static void PrintMatchesInMostFiles();
static void PrintMatchesInAllFiles();

#ifdef ASCIITRACE
static void
DebugDumpAll()
{
    /* for each file */
    t_File *Fp;


    for (Fp = Files; Fp; Fp = Fp->Next) {
	t_RMatchList *Mp;

	for (Mp = Fp->MatchList; Mp; Mp = Mp->Next) {
	    register t_Matches *mp;

	    for (mp = Mp->Matches; mp; mp = mp->Next) {
		printf("%s:\t%d %ld %ld %ld %s\n",
		    Fp->Name,
		    mp->WordsInPhrase,
		    mp->BlockInFile,
		    mp->WordInBlock,
		    Mp->FID,
		    Mp->Name
		);
		if (OneOnlyFlag) break;
	    }
	}
    }
}
#endif

static void
PrintResults()
{
#ifdef ASCIITRACE
    if (AsciiTrace > 6) {
	DebugDumpAll();
    }
#endif
#if 0
    if (Files->Next == (t_File *) NULL) {
	PrintMatchesInAnyFiles();
	return;
    }
#endif
    switch (RankType) {
    case e_Any:
	PrintMatchesInAnyFiles();
	return;
    case e_Most:
	PrintMatchesInMostFiles();
	return;
    case e_All:
	PrintMatchesInAllFiles();
	return;
    }
}

static void
PrintMatchesInMostFiles()
{
    t_File *Fp;
    int *MatchedPhrases;
    register int i;
    long TotalDistinctDocuments = 0;
    int MaxRank = 0;
    long TotalToPrint = 0;
    long Rank;

    /** First rank the documents by the number of phrases they match.
     **/

fprintf(stderr, "lqrank: quorum matching not yet available, sorry\n");
exit(1);

    /** Count the number of distinct documents **/
    for (i = 0; i <= FileCount; i++) {
	MatchedPhrases[i] = 0;
    }

    MatchedPhrases = (int *) emalloc(
	"lqrank.PrintMatchesInMostFiles.MatchedPhrases",
	sizeof(int) * TotalDistinctDocuments + 1
    );

    /* for each file */
    for (Fp = Files; Fp; Fp = Fp->Next) {
	t_RMatchList *Mp;

	for (Mp = Fp->MatchList; Mp; Mp = Mp->Next) {
	    register t_Matches *mp;

	    for (mp = Mp->Matches; mp; mp = mp->Next) {
		++TotalToPrint;
		MatchedPhrases[mp->WhichPhrase]++;
	    }
	}

	Fp->FileRank = 0;
	for (i = 0; i <= FileCount; i++) {
	    Fp->FileRank += (MatchedPhrases[i] > 0);
	    MatchedPhrases[i] = 0;
	}

	if (Fp->FileRank > MaxRank) {
	    MaxRank = Fp->FileRank;
	}

    }

    for (Rank = MaxRank; Rank >= 0; Rank--) {
	for (Fp = Files; Fp; Fp = Fp->Next) {
	    t_RMatchList *Mp;

printf("=== rank %f ===\n", (double) Rank);

	    if (Fp->FileRank < Rank) {
		printf("do %s later at rank %f\n", Fp->Name, Fp->FileRank);
		continue;
	    }

	    for (Mp = Fp->MatchList; Mp; Mp = Mp->Next) {
		register t_Matches *mp;


		for (mp = Mp->Matches; mp; mp = mp->Next) {
printf("file \"%s\" rank %f, phrase %d in \t", Fp->Name, Fp->FileRank, mp->WhichPhrase);
		    printf("%d %ld %ld %ld %s\n",
			mp->WordsInPhrase,
			mp->BlockInFile,
			mp->WordInBlock,
			Mp->FID,
			Mp->Name
		    );
		    if (OneOnlyFlag) break;
		}
	    }
	    Fp->FileRank = -1; /* so we don't print it again */
	}
    }
}

static void
PrintMatchesInAnyFiles()
{
    /* for each file */
    t_File *Fp;

    for (Fp = Files; Fp; Fp = Fp->Next) {
	t_RMatchList *Mp;

	for (Mp = Fp->MatchList; Mp; Mp = Mp->Next) {
	    register t_Matches *mp;

	    for (mp = Mp->Matches; mp; mp = mp->Next) {
		printf("%d %ld %ld %ld %s\n",
		    mp->WordsInPhrase,
		    mp->BlockInFile,
		    mp->WordInBlock,
		    Mp->FID,
		    Mp->Name
		);
		if (OneOnlyFlag) break;
	    }
	}
    }
}

static void
PrintMatchesInAllFiles()
{
    register t_RMatchList *Mp;

    /* For each document in the first list of matches: */
    for (Mp = Files->MatchList; Mp; Mp = Mp->Next) {
	char *Name;
	int Fail;
	t_File *Fp;

	Name = Mp->Name;
	Fail = 0;

	/* For each of the other lists */
	for (Fp = Files->Next; Fp; Fp = Fp->Next) {
	    register t_RMatchList *Mmp;
	    register int i;

	    /* does the list contain this document? */
	    for (Mmp = Fp->MatchList; Mmp; Mmp = Mmp->Next) {
		if ((i = STRCMP(Mmp->Name, Name)) >= 0) {
		    break;
		} 
	    }
	    /* we found it, or it isn't there*/

	    /* speedup for next time round: */
	    if (Fp != Files) {
		Fp->MatchList = Mmp;
	    }

	    if (!Mmp || i != 0) {
		Fail++;
		break;
	    }
	}

	if (Fail <= FailThreshold) {
	    if (OneOnlyFlag) {
		printf("%d %ld %ld %ld %s\n",
		    Mp->Matches->WordsInPhrase,
		    Mp->Matches->BlockInFile,
		    Mp->Matches->WordInBlock,
		    Mp->FID,
		    Mp->Name
		);
	    }

	    /* For _each_ of the lists */
	    for (Fp = Files; Fp; Fp = Fp->Next) {
		register t_RMatchList *Mp2;

		/* find the document (this loop will always zero times,
		 * because of the earlier optimisation involving altering
		 * Fp->MatchList to point to the first matching string.
		 * So we use strcmp() instead of STREQ.
		 */
		for (Mp2 = Fp->MatchList; Mp2; Mp2 = Mp2->Next) {
		    t_Matches *mp;
		    int j;
		    
		    j = STRCMP(Mp2->Name, Name);

		    if (j > 0) break;
		    if (j < 0) continue;

		    /* print out all the matches */
		    for (mp = Mp2->Matches; mp; mp = mp->Next) {
			if (!OneOnlyFlag) {
			    AddResult(
				mp->WordsInPhrase,
				mp->BlockInFile,
				mp->WordInBlock,
				Mp2->FID,
				Mp2->Name
			    );
			}
		    }
		}

		if (Fp != Files) {
		    Fp->MatchList = Mp2;
		}
	    }
	    AddResult(0, 0L, 0, 0L, (char *) 0); /* flush the cache */
	}
    }
    AddResult(0, 0L, 0, 0L, (char *) 0); /* flush the cache (paranoia) */
}
